from __future__ import annotations

import json
import tarfile
from typing import Any, Generator, Tuple

from ruamel import yaml

from conda_forge_metadata.libcfgraph import get_libcfgraph_artifact_data
from conda_forge_metadata.types import ArtifactData

VALID_BACKENDS = ("libcfgraph", "oci", "streamed")


def get_artifact_info_as_json(
    channel: str, subdir: str, artifact: str, backend: str = "libcfgraph"
) -> ArtifactData | None:
    """Get a blob of artifact data from the conda info directory.

    Parameters
    ----------
    channel : str
        The channel (e.g., "conda-forge").
    subdir : str
        The subdir for the artifact (e.g., "noarch", "linux-64", etc.).
    artifact : str
        The full artifact name with extension (e.g.,
        "21cmfast-3.0.2-py36h13dd421_0.tar.bz2").

    Returns
    -------
    info_blob : dict
        A dictionary of data. Possible keys are

            "metadata_version": the metadata version format
            "name": the package name
            "version": the package version
            "index": the info/index.json file contents
            "about": the info/about.json file contents
            "rendered_recipe": the fully rendered recipe at
                either info/recipe/meta.yaml or info/meta.yaml
                as a dict
            "raw_recipe": the template recipe as a string from
                info/recipe/meta.yaml.template - could be
                the rendered recipe as a string if no template was found
            "conda_build_config": the conda_build_config.yaml used for building
                the recipe at info/recipe/conda_build_config.yaml
            "files": a list of files in the recipe from info/files with
                elements ending in .pyc or .txt filtered out.
    """
    if backend == "libcfgraph":
        return get_libcfgraph_artifact_data(channel, subdir, artifact)
    elif backend == "oci":
        from conda_forge_metadata.oci import get_oci_artifact_data

        tar = get_oci_artifact_data(channel, subdir, artifact)
        if tar is not None:
            return info_json_from_tar_generator(tar)
    elif backend == "streamed":
        if artifact.endswith(".tar.bz2"):
            raise ValueError("streamed backend does not support .tar.bz2 artifacts")
        from conda_forge_metadata.streaming import get_streamed_artifact_data

        return info_json_from_tar_generator(
            get_streamed_artifact_data(channel, subdir, artifact)
        )
    else:
        raise ValueError(
            f"Unknown backend {backend!r}. Valid backends are {VALID_BACKENDS}."
        )


def info_json_from_tar_generator(
    tar_tuples: Generator[Tuple[tarfile.TarFile, tarfile.TarInfo], None, None],
) -> ArtifactData | None:
    # https://github.com/regro/libcflib/blob/062858e90af/libcflib/harvester.py#L14
    data = {
        "metadata_version": 1,
        "name": "",
        "version": "",
        "index": {},
        "about": {},
        "rendered_recipe": {},
        "raw_recipe": "",
        "conda_build_config": {},
        "files": [],
    }
    YAML = yaml.YAML(typ="safe")
    for tar, member in tar_tuples:
        if member.name.endswith("index.json"):
            index = json.loads(_extract_read(tar, member, default="{}"))
            data["name"] = index.get("name", "")
            data["version"] = index.get("version", "")
            data["index"] = index
        elif member.name.endswith("about.json"):
            data["about"] = json.loads(_extract_read(tar, member, default="{}"))
        elif member.name.endswith("conda_build_config.yaml"):
            data["conda_build_config"] = YAML.load(
                _extract_read(tar, member, default="{}")
            )
        elif member.name.endswith("files"):
            data["files"] = [
                f
                for f in _extract_read(tar, member, default="").splitlines()
                if not f.lower().endswith((".pyc", ".txt"))
            ]
        elif member.name.endswith("meta.yaml.template"):
            data["raw_recipe"] = _extract_read(tar, member, default="")
        elif member.name.endswith("meta.yaml"):
            x = _extract_read(tar, member, default="{}")
            if ("{{" in x or "{%" in x) and not data["raw_recipe"]:
                data["raw_recipe"] = x
            else:
                data["rendered_recipe"] = YAML.load(x)
    if data["name"]:
        return data  # type: ignore


def _extract_read(
    tar: tarfile.TarFile, member: tarfile.TarInfo, default: Any = None
) -> str:
    f = tar.extractfile(member)
    if f:
        return f.read().decode() or default
    return default



from conda_forge_metadata.feedstock_outputs import package_to_feedstock


def test_feedstock_outputs():
    assert package_to_feedstock("conda-forge-pinning") == ["conda-forge-pinning"]
    assert package_to_feedstock("tk") == ["tk"]
    assert "python" in package_to_feedstock("python")



from __future__ import annotations

from functools import lru_cache

import requests

from conda_forge_metadata.types import ArtifactData

_LIBCFGRAPH_INDEX = None


def _download_libcfgraph_index():
    global _LIBCFGRAPH_INDEX
    r = requests.get(
        "https://raw.githubusercontent.com/regro/libcfgraph"
        "/master/.file_listing_meta.json",
    )
    r.raise_for_status()
    n_files = r.json()["n_files"]
    _LIBCFGRAPH_INDEX = []
    for i in range(n_files):
        r = requests.get(
            "https://raw.githubusercontent.com/regro/libcfgraph"
            "/master/.file_listing_%d.json" % i,
        )
        r.raise_for_status()
        _LIBCFGRAPH_INDEX += r.json()


def get_libcfgraph_index() -> list[str]:
    """Get a list of all artifacts indexed by libcfgraph.

    Each element of the list looks like

        "artifacts/21cmfast/conda-forge/linux-64/21cmfast-3.0.2-py36h2e3f83d_0.json"

    for the conda-forge artifact

        linux-64/21cmfast-3.0.2-py36h2e3f83d_0.{tar.bz2,conda}
    """
    if _LIBCFGRAPH_INDEX is None:
        _download_libcfgraph_index()
    assert _LIBCFGRAPH_INDEX is not None
    return _LIBCFGRAPH_INDEX


@lru_cache(maxsize=1024)
def get_libcfgraph_artifact_data(
    channel: str, subdir: str, artifact: str
) -> ArtifactData | None:
    """Get a blob of artifact data from the conda info directory.

    Parameters
    ----------
    channel : str
        The channel (e.g., "conda-forge").
    subdir : str
        The subdir for the artifact (e.g., "noarch", "linux-64", etc.).
    artifact : str
        The full artifact name with extension (e.g.,
        "21cmfast-3.0.2-py36h13dd421_0.tar.bz2").

    Returns
    -------
    info_blob : dict
        A dictionary of data. Possible keys are

            "metadata_version": the metadata version format
            "name": the package name
            "version": the package version
            "index": the info/index.json file contents
            "about": the info/about.json file contents
            "rendered_recipe": the fully rendered recipe at
                either info/recipe/meta.yaml or info/meta.yaml
                as a dict
            "raw_recipe": the template recipe as a string from
                info/recipe/meta.yaml.template - could be
                the rendered recipe as a string if no template was found
            "conda_build_config": the conda_build_config.yaml used for building
                the recipe at info/recipe/conda_build_config.yaml
            "files": a list of files in the recipe from info/files with
                elements ending in .pyc or .txt filtered out.

        If the artifact is not indexed, it returns None.
    """
    # urls look like this:
    # https://raw.githubusercontent.com/regro/libcfgraph/master/
    #   artifacts/21cmfast/conda-forge/osx-64/21cmfast-3.0.2-py36h13dd421_0.json
    nm_parts = artifact.rsplit("-", 2)
    if artifact.endswith(".tar.bz2"):
        nm = artifact[: -len(".tar.bz2")]
    else:
        nm = artifact[: -len(".conda")]

    libcfgraph_path = "artifacts/" f"{nm_parts[0]}/{channel}/{subdir}/{nm}.json"
    if libcfgraph_path in get_libcfgraph_index():
        url = (
            "https://raw.githubusercontent.com/regro/libcfgraph/master/"
            + libcfgraph_path
        )
        r = requests.get(url)
        r.raise_for_status()
        return r.json()
    else:
        return None


@lru_cache(maxsize=1)
def _import_to_pkg_maps_num_letters() -> int:
    req = requests.get(
        "https://raw.githubusercontent.com/regro/libcfgraph/master"
        "/import_to_pkg_maps_meta.json"
    )
    req.raise_for_status()
    return int(req.json()["num_letters"])


@lru_cache(maxsize=128)
def _import_to_pkg_maps_cache(import_first_letters: str) -> dict[str, set[str]]:
    req = requests.get(
        f"https://raw.githubusercontent.com/regro/libcfgraph"
        f"/master/import_to_pkg_maps/{import_first_letters.lower()}.json"
    )
    req.raise_for_status()
    return {k: set(v["elements"]) for k, v in req.json().items()}


def _get_libcfgraph_pkgs_for_import(import_name: str) -> set[str] | None:
    num_letters = _import_to_pkg_maps_num_letters()
    fllt = import_name[: min(len(import_name), num_letters)]
    import_to_pkg_map = _import_to_pkg_maps_cache(fllt)
    return import_to_pkg_map.get(import_name, None)


def get_libcfgraph_pkgs_for_import(import_name: str) -> tuple[set[str] | None, str]:
    """Get a list of possible packages that supply a given import.

    **This data is approximate and may be wrong.**

    For a better guess, use the function

        `conda_forge_metadata.autotick_bot.map_import_to_package`

    which attempts to return the most likely supplying package.

    Parameters
    ----------
    import_name : str
        The name of the import.

    Returns
    -------
    packages : set
        A set of packages that possibly have the import.
        Will return `None` if the import was not found.
    found_import_name : str
        The import name found in the libcfgraph metadata. Only
        valid if `packages` is not None. This name will be the
        top-level import with all subpackages removed (e.g., foo.bar.baz
        will be returned as foo).
    """
    import_name = import_name.split(".")[0]
    supplying_pkgs = _get_libcfgraph_pkgs_for_import(import_name)
    return supplying_pkgs, import_name



import pytest

from conda_forge_metadata.artifact_info import info_json


@pytest.mark.parametrize("backend", info_json.VALID_BACKENDS)
def test_info_json_tar_bz2(backend: str):
    if backend == "streamed":
        pytest.xfail("streamed backend does not support .tar.bz2 artifacts")

    info = info_json.get_artifact_info_as_json(
        "conda-forge",
        "osx-64",
        "21cmfast-3.0.2-py36h13dd421_0.tar.bz2",
        backend=backend,
    )
    assert info is not None
    assert info["metadata_version"] == 1
    assert info["name"] == "21cmfast"
    assert info["version"] == "3.0.2"
    assert info["index"]["name"] == "21cmfast"
    assert info["index"]["version"] == "3.0.2"
    assert info["index"]["build"] == "py36h13dd421_0"
    assert info["index"]["subdir"] == "osx-64"
    assert "pyyaml" in info["index"]["depends"]
    assert info["about"]["conda_version"] == "4.8.4"
    assert info["rendered_recipe"]["package"]["name"] == "21cmfast"
    assert (
        info["rendered_recipe"]["source"]["sha256"]
        == "6e88960d134e98e4719343d853c63fc3c691438b57b2863f7834f07fae9eab4f"
    )
    assert info["raw_recipe"] is not None
    assert info["raw_recipe"].startswith('{% set name = "21cmFAST" %}')
    assert info["conda_build_config"]["CI"] == "azure"
    assert "bin/21cmfast" in info["files"]


@pytest.mark.parametrize("backend", info_json.VALID_BACKENDS)
def test_info_json_conda(backend: str):
    info = info_json.get_artifact_info_as_json(
        "conda-forge",
        "noarch",
        "abipy-0.9.6-pyhd8ed1ab_0.conda",
        backend=backend,
    )
    assert info is not None
    assert info["metadata_version"] == 1
    assert info["name"] == "abipy"
    assert info["version"] == "0.9.6"
    assert info["index"]["name"] == "abipy"
    assert info["index"]["version"] == "0.9.6"
    assert info["index"]["build"] == "pyhd8ed1ab_0"
    assert info["index"]["subdir"] == "noarch"
    assert "apscheduler" in info["index"]["depends"]
    assert info["about"]["conda_version"] == "23.3.1"
    assert info["rendered_recipe"]["package"]["name"] == "abipy"
    assert (
        info["rendered_recipe"]["source"]["sha256"]
        == "dc34c9179b9e53649353b30c1b37f0a36f5ea681fc541d60cafb3f4cf176cddf"
    )
    assert info["raw_recipe"] is not None
    assert info["raw_recipe"].startswith('{% set name = "abipy" %}')
    assert info["conda_build_config"]["CI"] == "azure"
    assert "site-packages/abipy/__init__.py" in info["files"]


@pytest.mark.parametrize("backend", info_json.VALID_BACKENDS)
def test_missing_conda_build_tar_bz2(backend: str):
    if backend == "streamed":
        pytest.xfail("streamed backend does not support .tar.bz2 artifacts")

    info = info_json.get_artifact_info_as_json(
        "conda-forge",
        "linux-64",
        "jinja2-2.10-py36_0.tar.bz2",
        backend=backend,
    )
    assert info is not None
    assert info["conda_build_config"] == {}



import tarfile
from logging import getLogger
from typing import Generator

from conda_oci_mirror.repo import PackageRepo

logger = getLogger(__name__)


def get_oci_artifact_data(
    channel: str,
    subdir: str,
    artifact: str,
    registry: str = "ghcr.io/channel-mirrors",
) -> "Generator[tuple[tarfile.TarFile, tarfile.TarInfo], None, None] | None":
    """Get a blob of artifact data from the conda info directory.

    Note this function might need token authentication to access the registry.
    Export the following environment variables:
    - ORAS_USER: the username to use for authentication
    - ORAS_PASS: the password/token to use for authentication

    Parameters
    ----------
    channel : str
        The channel (e.g., "conda-forge").
    subdir : str
        The subdir for the artifact (e.g., "noarch", "linux-64", etc.).
    artifact : str
        The full artifact name with extension (e.g.,
        "21cmfast-3.0.2-py36h13dd421_0.tar.bz2").
    registry : str
        The registry to use for the OCI repository.

    Returns
    -------
    info_blob : dict
        A dictionary of data. Possible keys are

            "metadata_version": the metadata version format
            "name": the package name
            "version": the package version
            "index": the info/index.json file contents
            "about": the info/about.json file contents
            "rendered_recipe": the fully rendered recipe at
                either info/recipe/meta.yaml or info/meta.yaml
                as a dict
            "raw_recipe": the template recipe as a string from
                info/recipe/meta.yaml.template - could be
                the rendered recipe as a string if no template was found
            "conda_build_config": the conda_build_config.yaml used for building
                the recipe at info/recipe/conda_build_config.yaml
            "files": a list of files in the recipe from info/files with
                elements ending in .pyc or .txt filtered out.

        If the artifact is not indexed, it returns None.
    """
    if artifact.endswith(".tar.bz2"):
        artifact = artifact[: -len(".tar.bz2")]
    elif artifact.endswith(".conda"):
        artifact = artifact[: -len(".conda")]
    else:
        raise ValueError(f"Artifact '{artifact}' is not a conda package")

    repo = PackageRepo(channel, subdir, None, registry=registry)

    parts = artifact.rsplit("-", 2)
    oci_name = f"{parts[0]}:{parts[1]}-{parts[2]}"
    try:
        tar = repo.get_info(oci_name)
    except ValueError as exc:
        logger.debug("Failed to get info for %s", oci_name, exc_info=exc)
        return None
    else:
        try:
            for member in tar:
                yield tar, member
        finally:
            tar.close()



from __future__ import annotations

from functools import lru_cache

import requests

from conda_forge_metadata.libcfgraph import get_libcfgraph_pkgs_for_import


@lru_cache(maxsize=1)
def _ranked_hubs_authorities() -> list[str]:
    req = requests.get(
        "https://raw.githubusercontent.com/regro/cf-graph-countyfair/"
        "master/ranked_hubs_authorities.json"
    )
    req.raise_for_status()
    return req.json()


def map_import_to_package(import_name: str) -> str:
    """Map an import name to the most likely package that has it.

    Parameters
    ----------
    import_name : str
        The name of the import.

    Returns
    -------
    pkg_name : str
        The name of the package.
    """
    supplying_pkgs, found_import_name = get_libcfgraph_pkgs_for_import(import_name)
    if supplying_pkgs is None:
        return found_import_name

    if found_import_name in supplying_pkgs:
        # heuristic that import scipy comes from scipy
        return found_import_name
    else:
        hubs_auths = _ranked_hubs_authorities()
        return next(
            iter(k for k in hubs_auths if k in supplying_pkgs),
            found_import_name,
        )
